@@ -1,4 +1,14 @@
-__version__ = "0.6.2"
+__version__ = "0.13.0"
+
+import sys
+
+if "pkg_resources" not in sys.modules:
+    # workaround for the following:
+    # https://github.com/benoitc/gunicorn/pull/2539
+    sys.modules["pkg_resources"] = object()  # type: ignore[assignment]
+    import aiohttp
+
+    del sys.modules["pkg_resources"]
 
 from . import read as read
 from . import share as share
