@@ -10,7 +10,7 @@ import time
 import urllib.parse
 from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, Tuple
 
-import xmltodict
+from .xml import etree
 
 if TYPE_CHECKING:
     from .path import AzurePath
@@ -29,27 +29,10 @@ def load_credentials() -> Dict[str, Any]:
     # AZURE_STORAGE_ACCOUNT_KEY is mentioned elsewhere on the internet
     for varname in ["AZURE_STORAGE_KEY", "AZURE_STORAGE_ACCOUNT_KEY"]:
         if varname in os.environ:
-            result = dict(storageAccountKey=os.environ[varname])
+            account = {}
             if "AZURE_STORAGE_ACCOUNT" in os.environ:
-                result["account"] = os.environ["AZURE_STORAGE_ACCOUNT"]
-            return result
-
-    if "AZURE_APPLICATION_CREDENTIALS" in os.environ:
-        creds_path = os.environ["AZURE_APPLICATION_CREDENTIALS"]
-        if not os.path.exists(creds_path):
-            raise RuntimeError(
-                f"Credentials not found at '{creds_path}' specified by environment variable "
-                "'AZURE_APPLICATION_CREDENTIALS'"
-            )
-        with open(creds_path) as f:
-            return json.load(f)
-
-    if "AZURE_CLIENT_ID" in os.environ:
-        return dict(
-            appId=os.environ["AZURE_CLIENT_ID"],
-            password=os.environ["AZURE_CLIENT_SECRET"],
-            tenant=os.environ["AZURE_TENANT_ID"],
-        )
+                account["account"] = os.environ["AZURE_STORAGE_ACCOUNT"]
+            return {"_azure_auth": "sakey", "storage_account_key": os.environ[varname], **account}
 
     if "AZURE_STORAGE_CONNECTION_STRING" in os.environ:
         connection_data = {}
@@ -58,39 +41,60 @@ def load_credentials() -> Dict[str, Any]:
         for part in os.environ["AZURE_STORAGE_CONNECTION_STRING"].split(";"):
             key, _, val = part.partition("=")
             connection_data[key.lower()] = val
-        return dict(
-            account=connection_data["accountname"], storageAccountKey=connection_data["accountkey"]
-        )
+        return {
+            "_azure_auth": "sakey",
+            "account": connection_data["accountname"],
+            "storage_account_key": connection_data["accountkey"],
+        }
+
+    if "AZURE_CLIENT_ID" in os.environ:
+        return {
+            "_azure_auth": "svcact",
+            "client_id": os.environ["AZURE_CLIENT_ID"],
+            "client_secret": os.environ["AZURE_CLIENT_SECRET"],
+            "tenant_id": os.environ["AZURE_TENANT_ID"],
+        }
 
     # look for a refresh token in the az command line credentials
-    # https://mikhail.io/2019/07/how-azure-cli-manages-access-tokens/
-    default_creds_path = os.path.expanduser("~/.azure/accessTokens.json")
-    if os.path.exists(default_creds_path):
-        with open(default_creds_path) as f:
+    # TODO: we could also try to use any found access tokens
+    msal_tokens_path = os.path.expanduser("~/.azure/msal_token_cache.json")
+    if os.path.exists(msal_tokens_path):
+        with open(msal_tokens_path) as f:
+            tokens = json.load(f)
+            for token in tokens.get("RefreshToken", {}).values():
+                if token["credential_type"] != "RefreshToken":
+                    continue
+                return {"_azure_auth": "refresh", "refresh_token": token["secret"]}
+
+    access_tokens_path = os.path.expanduser("~/.azure/accessTokens.json")
+    if os.path.exists(access_tokens_path):
+        with open(access_tokens_path) as f:
             tokens = json.load(f)
             best_token = None
             for token in tokens:
+                if "refreshToken" not in token:
+                    continue
+                creds = {"_azure_auth": "refresh", "refresh_token": token["refreshToken"]}
                 if best_token is None:
-                    best_token = token
+                    best_token = creds
                 else:
                     # expiresOn may be missing for tokens from service principals
                     if token.get("expiresOn", "") > best_token.get("expiresOn", ""):
-                        best_token = token
+                        best_token = creds
             if best_token is not None:
                 return best_token
 
     raise RuntimeError(
         """Azure credentials not found, please do one of the following:
 
-1) Log in with 'az login', blobfile will use your default credentials to lookup your storage
+1) Log in with 'az login', boostedblob will use your default credentials to lookup your storage
    account key
-2) Set the environment variable 'AZURE_STORAGE_KEY' to your storage account key which you can
-   find by following this guide:
-   https://docs.microsoft.com/en-us/azure/storage/common/storage-account-keys-manage
-3) Create an account with 'az ad sp create-for-rbac --name <name>' and set the
-   'AZURE_APPLICATION_CREDENTIALS' environment variable to the path of the output from that
-   command or individually set the 'AZURE_CLIENT_ID', 'AZURE_CLIENT_SECRET', and
+2) Create an account with 'az ad sp create-for-rbac --name <name>' and
+   individually set the 'AZURE_CLIENT_ID', 'AZURE_CLIENT_SECRET', and
    'AZURE_TENANT_ID' environment variables
+3) Set the environment variable 'AZURE_STORAGE_ACCOUNT_KEY' to your storage account key which you
+   can find by following this guide:
+   https://docs.microsoft.com/en-us/azure/storage/common/storage-account-keys-manage
 """
     )
 
@@ -109,7 +113,7 @@ def load_stored_subscription_ids() -> List[str]:
         # this file has a UTF-8 BOM
         profile = json.loads(f.read().decode("utf-8-sig"))
 
-    subscriptions = profile["subscriptions"]
+    subscriptions = profile.get("subscriptions", [])
     subscriptions.sort(key=lambda x: x["isDefault"], reverse=True)
     return [sub["id"] for sub in subscriptions]
 
@@ -120,24 +124,28 @@ async def get_access_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, f
     now = time.time()
     creds = load_credentials()
 
-    if "storageAccountKey" in creds:
+    if creds["_azure_auth"] == "sakey":
         if "account" in creds:
             if creds["account"] != account:
                 raise RuntimeError(
                     f"Provided storage account key for account '{creds['account']}' via "
                     f"environment variables, but needed credentials for account '{account}'"
                 )
-        auth = (SHARED_KEY, creds["storageAccountKey"])
+        auth = (SHARED_KEY, creds["storage_account_key"])
         if await can_access_account(account, container, auth):
             return (auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
         raise RuntimeError(
             f"Found storage account key, but it was unable to access storage account: '{account}'"
         )
 
-    if "refreshToken" in creds:
+    from .globals import config
+
+    if creds["_azure_auth"] == "refresh":
         # we have a refresh token, convert it into an access token for this account
         req = create_access_token_request(
-            creds=creds, scope=f"https://{account}.blob.core.windows.net/", success_codes=(200, 400)
+            creds=creds,
+            scope=f"https://{account}.blob.core.windows.net/.default",
+            success_codes=(200, 400),
         )
 
         async with req.execute() as resp:
@@ -166,15 +174,16 @@ async def get_access_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, f
             return (auth, now + float(result["expires_in"]))
 
         # it didn't work, fall back to getting the storage keys
-        storage_account_key_auth = await get_storage_account_key(
-            account=account, creds=creds, container_hint=container
-        )
-        if storage_account_key_auth is not None:
-            return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
+        if config.storage_account_key_fallback:
+            storage_account_key_auth = await get_storage_account_key(
+                account=account, creds=creds, container_hint=container
+            )
+            if storage_account_key_auth is not None:
+                return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
 
-    if "appId" in creds:
+    if creds["_azure_auth"] == "svcact":
         # we have a service principal, get an oauth token
-        req = create_access_token_request(creds=creds, scope="https://storage.azure.com/")
+        req = create_access_token_request(creds=creds, scope="https://storage.azure.com/.default")
 
         async with req.execute() as resp:
             result = await resp.json()
@@ -183,11 +192,12 @@ async def get_access_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, f
             return (auth, now + float(result["expires_in"]))
 
         # it didn't work, fall back to getting the storage keys
-        storage_account_key_auth = await get_storage_account_key(
-            account=account, creds=creds, container_hint=container
-        )
-        if storage_account_key_auth is not None:
-            return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
+        if config.storage_account_key_fallback:
+            storage_account_key_auth = await get_storage_account_key(
+                account=account, creds=creds, container_hint=container
+            )
+            if storage_account_key_auth is not None:
+                return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
 
     raise RuntimeError(
         f"Could not find any credentials that grant access to storage account: '{account}'"
@@ -212,14 +222,14 @@ async def can_access_account(account: str, container: Optional[str], auth: Tuple
                 return False
             data = await resp.read()
 
-        out = xmltodict.parse(data)
-        if out["EnumerationResults"]["Containers"] is None:
+        result = etree.fromstring(data)
+        container = result.findtext("Containers/Container/Name")
+
+        if container is None:
             # there are no containers in this storage account
             # we can't test if we can access this storage account or not, so presume we can
             return True
-
-        # then also test that we can list a container. this is perhaps unnecessary...
-        container = out["EnumerationResults"]["Containers"]["Container"]["Name"]
+        # then also test that we can list that container. this is perhaps unnecessary...
 
     # https://myaccount.blob.core.windows.net/mycontainer?restype=container&comp=list
     req = Request(
@@ -238,32 +248,27 @@ def create_access_token_request(
 ) -> Request:
     from .request import Request
 
-    if "refreshToken" in creds:
-        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v1-protocols-oauth-code#refreshing-the-access-tokens
+    if creds["_azure_auth"] == "refresh":
+        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-auth-code-flow#refresh-the-access-token
         data = {
             "grant_type": "refresh_token",
-            "refresh_token": creds["refreshToken"],
-            "resource": scope,
+            "refresh_token": creds["refresh_token"],
+            "scope": scope,
         }
-        tenant = "common"
-    else:
-        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v1-oauth2-client-creds-grant-flow#request-an-access-token
-        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v1-protocols-oauth-code
-        # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-azure-active-directory#use-oauth-access-tokens-for-authentication
-        # https://docs.microsoft.com/en-us/rest/api/azure/
-        # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-azure-active-directory
-        # az ad sp create-for-rbac --name <name>
-        # az account list
-        # az role assignment create --role "Storage Blob Data Contributor" --assignee <appid> --scope "/subscriptions/<account id>"
+        tenant_id = "common"
+    elif creds["_azure_auth"] == "svcact":
+        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-client-creds-grant-flow#first-case-access-token-request-with-a-shared-secret
         data = {
             "grant_type": "client_credentials",
-            "client_id": creds["appId"],
-            "client_secret": creds["password"],
-            "resource": scope,
+            "client_id": creds["client_id"],
+            "client_secret": creds["client_secret"],
+            "scope": scope,
         }
-        tenant = creds["tenant"]
+        tenant_id = creds["tenant_id"]
+    else:
+        raise AssertionError
     return Request(
-        url=f"https://login.microsoftonline.com/{tenant}/oauth2/token",
+        url=f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
         method="POST",
         headers={"Content-Type": "application/x-www-form-urlencoded"},
         data=urllib.parse.urlencode(data).encode("utf8"),
@@ -280,18 +285,41 @@ async def get_storage_account_id_with_subscription(
         method="GET",
         url=f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Storage/storageAccounts",
         params={"api-version": "2019-04-01"},
-        success_codes=(200, 401, 403),
+        success_codes=(200, 401, 403, 429),
     )
-    req = await azurify_request(req, auth=auth)
 
-    async with req.execute() as resp:
-        if resp.status in (401, 403):
-            # we aren't allowed to query this for this subscription, skip it
-            return None
-        out = await resp.json()
+    while True:
+        req = await azurify_request(req, auth=auth)
+        async with req.execute() as resp:
+            if resp.status in (401, 403):
+                # we aren't allowed to query this for this subscription, skip it
+                return None
+            if resp.status == 429:
+                # these API endpoints have really low rate limits
+                # note this error message only makes sense because the only code path that
+                # currently calls get_storage_account_id_with_subscription is the storage
+                # account key fallback
+                raise RuntimeError(
+                    "You do not have the correct IAM roles to access this location. "
+                    "Check that you have Storage Blob Data Reader or Storage Blob Data Contributor "
+                    "roles. You can run "
+                    f"`az storage container list --auth-mode login --account-name {account}` "
+                    "to confirm the missing role.\n"
+                    "In this situation, boostedblob is capable of falling back to accessing the "
+                    "storage account keys, but Azure has really low rate limits on these queries, "
+                    "which are currently preventing us from doing so."
+                )
+            result = await resp.json()
 
-    # search for the storage account
-    return next((obj["id"] for obj in out["value"] if obj["name"] == account), None)
+        # search for the storage account
+        account_id = next((obj["id"] for obj in result["value"] if obj["name"] == account), None)
+        if account_id is not None:
+            return account_id
+
+        if "nextLink" not in result:
+            break
+        req = Request(method="GET", url=result["nextLink"], success_codes=(200, 401, 403))
+    return None
 
 
 async def get_storage_account_id(account: str, auth: Tuple[str, str]) -> Optional[str]:
@@ -310,17 +338,23 @@ async def get_storage_account_id(account: str, auth: Tuple[str, str]) -> Optiona
         url="https://management.azure.com/subscriptions",
         params={"api-version": "2020-01-01"},
     )
-    req = await azurify_request(req, auth=auth)
-    async with req.execute() as resp:
-        result = await resp.json()
-    subscription_ids = {item["subscriptionId"] for item in result["value"]}
 
-    for subscription_id in subscription_ids - set(stored_subscription_ids):
-        storage_account_id = await get_storage_account_id_with_subscription(
-            subscription_id, account, auth
-        )
-        if storage_account_id:
-            return storage_account_id
+    while True:
+        req = await azurify_request(req, auth=auth)
+        async with req.execute() as resp:
+            result = await resp.json()
+        subscription_ids = {item["subscriptionId"] for item in result["value"]}
+
+        for subscription_id in subscription_ids - set(stored_subscription_ids):
+            storage_account_id = await get_storage_account_id_with_subscription(
+                subscription_id, account, auth
+            )
+            if storage_account_id:
+                return storage_account_id
+
+        if "nextLink" not in result:
+            break
+        req = Request(method="GET", url=result["nextLink"])
     return None
 
 
@@ -330,7 +364,7 @@ async def get_storage_account_key(
     from .request import Request, azurify_request
 
     # get an access token for the management service
-    req = create_access_token_request(creds=creds, scope="https://management.azure.com/")
+    req = create_access_token_request(creds=creds, scope="https://management.azure.com/.default")
 
     async with req.execute() as resp:
         result = await resp.json()
@@ -355,7 +389,8 @@ async def get_storage_account_key(
             if await can_access_account(account, container_hint, storage_key_auth):
                 return storage_key_auth
             raise RuntimeError(
-                f"Found storage account key, but it was unable to access storage account: '{account}'"
+                "Found storage account key, but it was unable to access "
+                f"storage account: '{account}'"
             )
     raise RuntimeError(
         f"Storage account was found, but storage account keys were missing: '{account}'"
@@ -427,7 +462,10 @@ async def get_sas_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, floa
         cmd = (
             f"az storage container list --auth-mode login --account-name {account}"
             if container is None
-            else f"az storage blob list --auth-mode login --account-name {account} --container {container}"
+            else (
+                f"az storage blob list --auth-mode login --account-name {account} "
+                f"--container {container}"
+            )
         )
         raise RuntimeError(
             "Only OAuth tokens can be used to get SAS tokens. You should set the "
@@ -459,8 +497,9 @@ async def get_sas_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, floa
             )
         data = await resp.read()
 
-    out = xmltodict.parse(data)
-    return out["UserDelegationKey"], time.time() + AZURE_SAS_TOKEN_EXPIRATION_SECONDS
+    result = etree.fromstring(data)
+    user_delegation_key = {el.tag: el.text for el in result}
+    return user_delegation_key, time.time() + AZURE_SAS_TOKEN_EXPIRATION_SECONDS
 
 
 async def generate_signed_url(path: AzurePath) -> Tuple[str, datetime.datetime]:
