@@ -1,11 +1,13 @@
 import argparse
 import asyncio
+import datetime
 import functools
 import os
+import shlex
 import subprocess
 import sys
 import tempfile
-from typing import Any, AsyncIterator, Awaitable, Callable, Dict, List, TypeVar
+from typing import Any, AsyncIterator, Callable, Coroutine, Dict, List, Optional, TypeVar, cast
 
 import boostedblob as bbb
 
@@ -13,14 +15,15 @@ F = TypeVar("F", bound=Callable[..., Any])
 T = TypeVar("T")
 
 
-def syncify(fn: Callable[..., Awaitable[T]]) -> Callable[..., T]:
+def syncify(fn: Callable[..., Coroutine[T, None, None]]) -> Callable[..., T]:
     @functools.wraps(fn)
     def wrapper(*args: Any, **kwargs: Any) -> Any:
         try:
-            if sys.version_info < (3, 9):
-                # waiting for https://github.com/MagicStack/uvloop/issues/349
-                import uvloop
+            import uvloop
 
+            if sys.version_info < (3, 9) or (
+                tuple(map(int, cast(Any, uvloop).__version__.split("."))) >= (0, 15, 0)
+            ):
                 uvloop.install()
         except ImportError:
             pass
@@ -30,17 +33,53 @@ def syncify(fn: Callable[..., Awaitable[T]]) -> Callable[..., T]:
 
 
 def sync_with_session(fn: F) -> F:
-    return syncify(bbb.ensure_session(fn))  # type: ignore
+    return syncify(bbb.ensure_session(fn))  # type: ignore[return-value]
 
 
-DEFAULT_CONCURRENCY = int(os.environ.get("BBB_DEFAULT_CONCURRENCY", 100))
+DEFAULT_CONCURRENCY = int(os.environ.get("BBB_DEFAULT_CONCURRENCY", 32))
 
 
 def is_glob(path: str) -> bool:
     return "*" in path
 
 
-async def print_long(it: AsyncIterator[bbb.listing.DirEntry], human_readable: bool) -> None:
+def glob_parent(path: str) -> str:
+    return path.split("*", 1)[0].rsplit("/", 1)[0]
+
+
+def format_size(num: float, suffix: str = "B") -> str:
+    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
+        if abs(num) < 1024.0:
+            unit += suffix
+            return f"{num:.1f} {unit:<3}"
+        num /= 1024.0
+    return f"{num:.1f} Yi{suffix}"
+
+
+def format_path_relative(path: bbb.BasePath, relative_to: Optional[bbb.BasePath]) -> str:
+    if relative_to is None:
+        return str(path)
+    if relative_to == path:
+        return path.name
+    return path.relative_to(relative_to)
+
+
+def format_long_entry(
+    entry: bbb.listing.DirEntry, human_readable: bool, relative_to: Optional[bbb.BasePath]
+) -> str:
+    size = (
+        (format_size(entry.stat.size) if human_readable else entry.stat.size) if entry.stat else ""
+    )
+    mtime = datetime.datetime.fromtimestamp(int(entry.stat.mtime)).isoformat() if entry.stat else ""
+    path = format_path_relative(entry.path, relative_to)
+    return f"{size:>12}  {mtime:19}  {path}"
+
+
+async def print_long(
+    it: AsyncIterator[bbb.listing.DirEntry],
+    human_readable: bool,
+    relative_to: Optional[bbb.BasePath],
+) -> None:
     total = 0
     num_files = 0
     async for entry in it:
@@ -48,60 +87,143 @@ async def print_long(it: AsyncIterator[bbb.listing.DirEntry], human_readable: bo
             num_files += 1
         if entry.stat:
             total += entry.stat.size
-        print(entry.format(human_readable=human_readable))
+        print(format_long_entry(entry, human_readable=human_readable, relative_to=relative_to))
     if human_readable:
-        human_total = bbb.listing.format_size(total).strip()
+        human_total = format_size(total).strip()
         print(f"Listed {num_files} files summing to {total} bytes ({human_total})")
 
 
 @sync_with_session
-async def ls(path: str, long: bool = False, machine: bool = False) -> None:
+async def ls(path: str, long: bool = False, machine: bool = False, relative: bool = False) -> None:
     path_obj = bbb.BasePath.from_str(path)
+    if isinstance(path_obj, bbb.LocalPath):
+        path_obj = path_obj.abspath()
+
     if is_glob(path):
+        relative_to = bbb.BasePath.from_str(glob_parent(path)) if relative else None
         it = bbb.listing.glob_scandir(path_obj)
         if long:
-            await print_long(it, human_readable=not machine)
+            await print_long(it, human_readable=not machine, relative_to=relative_to)
         else:
             async for entry in it:
-                print(entry.path)
+                print(format_path_relative(entry.path, relative_to))
         return
 
+    relative_to = path_obj if relative else None
     try:
         if long:
-            await print_long(bbb.scandir(path_obj), human_readable=not machine)
+            await print_long(
+                bbb.scandir(path_obj),
+                human_readable=not machine,
+                relative_to=path_obj if relative else None,
+            )
         else:
             async for p in bbb.listdir(path_obj):
-                print(p)
+                print(format_path_relative(p, relative_to))
     except NotADirectoryError:
         if long:
             stat = await bbb.stat(path_obj)
             entry = bbb.listing.DirEntry.from_path_stat(path_obj, stat)
-            print(entry.format(human_readable=not machine))
+            print(format_long_entry(entry, human_readable=not machine, relative_to=relative_to))
         else:
-            print(path_obj)
+            print(format_path_relative(path_obj, relative_to))
 
 
 @sync_with_session
-async def lstree(path: str, long: bool = False, machine: bool = False) -> None:
+async def lstree(
+    path: str, long: bool = False, machine: bool = False, relative: bool = False
+) -> None:
+    path_obj = bbb.BasePath.from_str(path)
+    if isinstance(path_obj, bbb.LocalPath):
+        path_obj = path_obj.abspath()
+    relative_to = path_obj if relative else None
+
     try:
         if long:
-            await print_long(bbb.scantree(path), human_readable=not machine)
+            await print_long(
+                bbb.scantree(path), human_readable=not machine, relative_to=relative_to
+            )
         else:
             async for p in bbb.listtree(path):
-                print(p)
+                print(format_path_relative(p, relative_to))
     except NotADirectoryError:
-        path_obj = bbb.BasePath.from_str(path)
         if long:
             stat = await bbb.stat(path_obj)
             entry = bbb.listing.DirEntry.from_path_stat(path_obj, stat)
-            print(entry.format(human_readable=not machine))
+            print(format_long_entry(entry, human_readable=not machine, relative_to=relative_to))
         else:
-            print(path_obj)
+            print(format_path_relative(path_obj, relative_to))
+
+
+@sync_with_session
+async def _dud1(path: str) -> None:
+    sizes: Dict[bbb.BasePath, int] = {}
+
+    total_count = 0
+
+    async def _dud0(entry: bbb.listing.DirEntry) -> None:
+        nonlocal total_count
+        if entry.is_file:
+            sizes[entry.path] = entry.stat.size if entry.stat else 0
+            total_count += 1
+            return
+        sizes[entry.path] = 0
+        async for e in bbb.scantree(entry.path):
+            if e.stat:
+                sizes[entry.path] += e.stat.size
+                total_count += 1
+
+    def print_sizes() -> int:
+        sorted_sizes = sorted(sizes.items(), key=lambda x: x[1])
+        for subpath, size in sorted_sizes:
+            print(f"{format_size(size).strip():>12}  {subpath}")
+        return len(sorted_sizes)
+
+    def clear_lines(num_lines: int) -> None:
+        erase_in_line = "\x1b[2K"
+        cursor_up = "\x1b[1A"
+        clear_scrollback = "\x1b[3J"
+        print("\r" + f"{cursor_up}{erase_in_line}" * num_lines, end="")
+        print(clear_scrollback, end="")
+
+    def print_spinner(message: str, _pos: Any = [0]) -> None:
+        clocks = ["🕛", "🕐", "🕑", "🕒", "🕓", "🕔", "🕕", "🕖", "🕗", "🕘", "🕙", "🕚"]
+        print(f"{clocks[_pos[0]]} {message}")
+        _pos[0] = (_pos[0] + 1) % len(clocks)
+
+    loop = asyncio.get_running_loop()
+    finished = loop.create_future()
+
+    async def live_update() -> None:
+        last_printed_len = 0
+        while True:
+            clear_lines(last_printed_len)
+            last_printed_len = print_sizes()
+            print_spinner(f"Summed {total_count} files so far...")
+            last_printed_len += 1
+            try:
+                await asyncio.wait_for(asyncio.shield(finished), 0.25)
+                break
+            except asyncio.TimeoutError:
+                pass
+        clear_lines(last_printed_len)
+
+    printer = asyncio.create_task(live_update())
+
+    async with bbb.BoostExecutor(DEFAULT_CONCURRENCY) as executor:
+        await bbb.boost.consume(executor.map_unordered(_dud0, executor.eagerise(bbb.scandir(path))))
+
+    finished.set_result(None)
+    await printer
+
+    print_sizes()
+    total_size = format_size(sum(sizes.values())).strip()
+    print(f"Listed {total_count} files summing to {total_size}")
 
 
 @sync_with_session
 async def cat(path: str, concurrency: int = DEFAULT_CONCURRENCY) -> None:
-    loop = asyncio.get_event_loop()
+    loop = asyncio.get_running_loop()
     async with bbb.BoostExecutor(concurrency) as executor:
         stream = await bbb.read.read_stream(path, executor)
         async for data in bbb.boost.iter_underlying(stream):
@@ -185,6 +307,16 @@ async def rmtree(path: str, quiet: bool = False, concurrency: int = DEFAULT_CONC
             await bbb.rmtree(path_obj, executor)
 
 
+@sync_with_session
+async def _xxx_recoverprefix(
+    prefix: str, restore_dt: str, dry_run: bool = True, concurrency: int = DEFAULT_CONCURRENCY
+) -> None:
+    from ._recover import _recoverprefix
+
+    async with bbb.BoostExecutor(concurrency) as executor:
+        await _recoverprefix(prefix, restore_dt, executor=executor, dry_run=dry_run)
+
+
 @sync_with_session
 async def share(path: str) -> None:
     url, expiration = await bbb.share.get_url(path)
@@ -198,6 +330,7 @@ async def sync(
     src: str,
     dst: str,
     delete: bool = False,
+    exclude: Optional[str] = None,
     quiet: bool = False,
     concurrency: int = DEFAULT_CONCURRENCY,
 ) -> None:
@@ -208,26 +341,39 @@ async def sync(
     if not src_is_dirlike:
         raise ValueError(f"{src_obj} is not a directory")
     async with bbb.BoostExecutor(concurrency) as executor:
-        async for p in bbb.sync(src_obj, dst_obj, executor, delete=delete):
+        async for p in bbb.sync(src_obj, dst_obj, executor, delete=delete, exclude=exclude):
             if not quiet:
                 print(p)
 
 
 @sync_with_session
-async def edit(path: str) -> None:
+async def edit(path: str, read_only: bool = False) -> None:
     with tempfile.TemporaryDirectory() as tmpdir:
         path_obj = bbb.BasePath.from_str(path)
         local = bbb.LocalPath(tmpdir) / path_obj.name
         async with bbb.BoostExecutor(DEFAULT_CONCURRENCY) as executor:
-            await bbb.copyfile(path_obj, local, executor)
+            try:
+                await bbb.copyfile(path_obj, local, executor)
+            except FileNotFoundError:
+                print("File not found, creating new file...")
+                with open(local, "w"):
+                    pass
             pre_stat = await bbb.stat(local)
-            subprocess.check_call([os.environ.get("EDITOR", "vi"), local])
+            editor = shlex.split(os.environ.get("EDITOR", "vi"))
+            subprocess.check_call([*editor, local])
             post_stat = await bbb.stat(local)
             if pre_stat != post_stat:
-                await bbb.copyfile(local, path_obj, executor, overwrite=True)
-                print(f"Updated {path_obj}")
+                if read_only:
+                    raise RuntimeError(
+                        "File was edited, but read-only mode is enabled. "
+                        "Edits were not persisted."
+                    )
+                else:
+                    await bbb.copyfile(local, path_obj, executor, overwrite=True)
+                    print(f"Updated {path_obj}")
             else:
-                print("File unmodified, skipping reupload...")
+                if not read_only:
+                    print("File unmodified, skipping reupload...")
 
 
 def complete_init(shell: str) -> None:
@@ -285,8 +431,11 @@ async def complete_command(shell: str, index: int, partial_command: List[str]) -
             # bash won't let you complete before a colon without setting COMP_WORDBREAKS globally
             # instead, we just make sure our completion matches the part before the colon
             # to do that, we special case az:// paths
-            if isinstance(entry.path, bbb.AzurePath) and path_to_complete.startswith("az://"):
-                path_str = entry.path.to_az_str()
+            if isinstance(entry.path, bbb.AzurePath):
+                if path_to_complete.startswith("az://"):
+                    path_str = entry.path.to_az_str()
+                else:
+                    path_str = entry.path.to_https_str()
             else:
                 path_str = str(entry.path)
 
@@ -328,8 +477,12 @@ $ bbb ls -l .
 To check whether a file exists:
 $ bbb ls schrodingers_file.txt
 
+To show relative paths:
+$ bbb ls --relative .
+
 Aliases:
 bbb ll == bbb ls -l
+bbb ll -s == bbb ls --long --relative
 """
     lstree_desc = """\
 `bbb lstree` lists all files present anywhere in the given directory tree.
@@ -345,9 +498,13 @@ $ bbb lstree -l gs://my_bucket/
 To make output more easily machine readable:
 $ bbb lstree --machine .
 
+To show relative paths:
+$ bbb lstree --relative .
+
 Aliases:
 bbb lsr == bbb lstree
 bbb llr == bbb lstree -l == bbb lsr -l == bbb du
+bbb lsr -s == bbb lstree --relative
 """
     cat_desc = """\
 `bbb cat` copies the contents of a given file to stdout.
@@ -413,7 +570,11 @@ $ bbb share gs://bucket/frogs.txt
 bbb will change the destination so that it better mirrors the source.
 Specifically, bbb will copy over or replace files in the destination that are
 missing or have changed in the source. If --delete is specified, bbb will delete
-files in the destination that are not present in the source.
+files in the destination that are not present in the source. If --exclude (-x)
+is specified, bbb will not copy or delete any files whose relative path matches
+the given regex at any position (i.e. using the semantics of Python's
+re.search). Consider using regex anchors (^$) to ensure you're excluding exactly
+what you intend.
 
 Example:
 $ bbb sync gs://tmp/boostedblob boostedblob
@@ -421,6 +582,9 @@ $ bbb sync gs://tmp/boostedblob boostedblob
 Using --delete and re-syncing will delete spurious_file.txt:
 $ touch boostedblob/spurious_file.txt
 $ bbb sync --delete gs://tmp/boostedblob boostedblob
+
+Sync all files except those with .txt extension
+$ bbb sync gs://tmp/boostedblob boostedblob -x '\\.txt$'
 """
     complete_desc = """\
 To enable tab completion for bash, add the following to your bashrc:
@@ -442,10 +606,9 @@ eval "$(bbb complete init zsh)"
         "-l", "--long", action="store_true", help="List information about each file"
     )
     subparser.add_argument(
-        "--machine",
-        action="store_true",
-        help="Make output more easily machine readable",
+        "--machine", action="store_true", help="Make output more easily machine readable"
     )
+    subparser.add_argument("-s", "--relative", action="store_true", help="Show relative paths")
 
     subparser = subparsers.add_parser(
         "ll", formatter_class=argparse.RawDescriptionHelpFormatter, description=ls_desc
@@ -453,10 +616,9 @@ eval "$(bbb complete init zsh)"
     subparser.set_defaults(command=functools.partial(ls, long=True))
     subparser.add_argument("path", help="Path of directory to list")
     subparser.add_argument(
-        "--machine",
-        action="store_true",
-        help="Make output more easily machine readable",
+        "--machine", action="store_true", help="Make output more easily machine readable"
     )
+    subparser.add_argument("-s", "--relative", action="store_true", help="Show relative paths")
 
     subparser = subparsers.add_parser(
         "lstree",
@@ -471,10 +633,9 @@ eval "$(bbb complete init zsh)"
         "-l", "--long", action="store_true", help="List information about each file"
     )
     subparser.add_argument(
-        "--machine",
-        action="store_true",
-        help="Make output more easily machine readable",
+        "--machine", action="store_true", help="Make output more easily machine readable"
     )
+    subparser.add_argument("-s", "--relative", action="store_true", help="Show relative paths")
 
     subparser = subparsers.add_parser(
         "llr",
@@ -485,10 +646,13 @@ eval "$(bbb complete init zsh)"
     subparser.set_defaults(command=functools.partial(lstree, long=True))
     subparser.add_argument("path", help="Root of directory tree to list")
     subparser.add_argument(
-        "--machine",
-        action="store_true",
-        help="Make output more easily machine readable",
+        "--machine", action="store_true", help="Make output more easily machine readable"
     )
+    subparser.add_argument("-s", "--relative", action="store_true", help="Show relative paths")
+
+    subparser = subparsers.add_parser("_dud1")
+    subparser.set_defaults(command=_dud1)
+    subparser.add_argument("path", help="Path of directory to list")
 
     subparser = subparsers.add_parser(
         "cat",
@@ -533,6 +697,7 @@ eval "$(bbb complete init zsh)"
     )
     subparser.set_defaults(command=edit)
     subparser.add_argument("path")
+    subparser.add_argument("-r", "--read-only", action="store_true", help="Open file read-only")
 
     subparser = subparsers.add_parser(
         "rm",
@@ -578,9 +743,23 @@ eval "$(bbb complete init zsh)"
     subparser.add_argument(
         "--delete", action="store_true", help="Delete destination files that don't exist in source"
     )
+    subparser.add_argument(
+        "-x",
+        "--exclude",
+        help=(
+            "Only copy or delete files whose relative path don't match this Python regular "
+            "expression at any position"
+        ),
+    )
     subparser.add_argument("-q", "--quiet", action="store_true")
     subparser.add_argument("--concurrency", **concurrency_kwargs)
 
+    subparser = subparsers.add_parser("_xrp")
+    subparser.set_defaults(command=_xxx_recoverprefix)
+    subparser.add_argument("prefix", help="Prefix of globs to recover")
+    subparser.add_argument("restore_dt", help="GMT datetime to recover to")
+    subparser.add_argument("--dry-run", type=lambda x: not x or x[0].lower() != "f", default=True)
+
     subparser = subparsers.add_parser(
         "complete",
         help="Tab complete a command",
