@@ -3,11 +3,12 @@ from __future__ import annotations
 import asyncio
 import base64
 import datetime
+import email.utils
 import functools
 import os
 import urllib.parse
 from dataclasses import dataclass
-from typing import Any, Callable, Mapping, NamedTuple, Optional, TypeVar, Union
+from typing import Any, Callable, Mapping, Optional, TypeVar, Union
 
 from .request import Request, azure_page_iterator, azurify_request, googlify_request
 
@@ -46,7 +47,7 @@ class BasePath:
 
     def is_relative_to(self: T, other: T) -> bool:
         try:
-            self.relative_to(other)  # type: ignore
+            self.relative_to(other)  # type: ignore[attr-defined]
             return True
         except ValueError:
             return False
@@ -78,7 +79,10 @@ class LocalPath(BasePath):
 
     @property
     def parent(self) -> LocalPath:
-        return LocalPath(os.path.dirname(self.path) or ".")
+        parent = os.path.dirname(_strip_slash(self.path))
+        if not parent:
+            parent = "/" if os.path.isabs(self.path) else "."
+        return LocalPath(parent)
 
     def relative_to(self, other: LocalPath) -> str:
         if not isinstance(other, LocalPath):
@@ -97,6 +101,9 @@ class LocalPath(BasePath):
     def ensure_directory_like(self) -> LocalPath:
         return self if self.is_directory_like() else LocalPath(self.path + "/")
 
+    def abspath(self) -> LocalPath:
+        return LocalPath(os.path.abspath(self.path))
+
     def __truediv__(self, relative_path: str) -> LocalPath:
         return LocalPath(os.path.join(self.path, relative_path))
 
@@ -123,17 +130,17 @@ class AzurePath(CloudPath):
         if parsed_url.scheme == "https":
             account, host = parsed_url.netloc.split(".", maxsplit=1)
             if host != "blob.core.windows.net":
-                raise ValueError(f"Invalid URL '{url}'")
+                raise ValueError(f"Invalid URL '{url}'; unexpected host '{host}'")
         elif parsed_url.scheme == "az":
             account = parsed_url.netloc
         else:
-            raise ValueError(f"Invalid URL '{url}'")
+            raise ValueError(f"Invalid URL '{url}'; expected 'https' or 'az' scheme")
 
-        parts = parsed_url.path.split("/", maxsplit=2)
-        if parts[0]:
-            raise ValueError(f"Invalid URL '{url}'")
-        container = parts[1] if len(parts) >= 2 else ""
-        return AzurePath(account=account, container=container, blob="/".join(parts[2:]))
+        # split the unparsed URL
+        parts = url.split("/", maxsplit=4)
+        container = parts[3] if len(parts) >= 4 else ""
+        blob = parts[4] if len(parts) >= 5 else ""
+        return AzurePath(account=account, container=container, blob=blob)
 
     @property
     def name(self) -> str:
@@ -141,7 +148,7 @@ class AzurePath(CloudPath):
 
     @property
     def parent(self) -> AzurePath:
-        return AzurePath(self.account, self.container, os.path.dirname(self.blob))
+        return AzurePath(self.account, self.container, os.path.dirname(_strip_slash(self.blob)))
 
     def relative_to(self, other: AzurePath) -> str:
         if not isinstance(other, AzurePath):
@@ -169,6 +176,9 @@ class AzurePath(CloudPath):
         return AzurePath(self.account, self.container, os.path.join(self.blob, relative_path))
 
     def __str__(self) -> str:
+        return self.to_az_str()
+
+    def to_https_str(self) -> str:
         return f"https://{self.account}.blob.core.windows.net/{self.container}/{self.blob}"
 
     def to_az_str(self) -> str:
@@ -187,9 +197,12 @@ class GooglePath(CloudPath):
     def from_str(url: str) -> GooglePath:
         parsed_url = urllib.parse.urlparse(url)
         if parsed_url.scheme != "gs":
-            raise ValueError(f"Invalid URL '{url}'")
+            raise ValueError(f"Invalid URL '{url}'; expected 'gs' scheme")
 
-        return GooglePath(bucket=parsed_url.netloc, blob=parsed_url.path[1:])
+        # split the unparsed URL
+        parts = url.split("/", maxsplit=3)
+        blob = parts[3] if len(parts) >= 4 else ""
+        return GooglePath(bucket=parsed_url.netloc, blob=blob)
 
     @property
     def name(self) -> str:
@@ -197,7 +210,7 @@ class GooglePath(CloudPath):
 
     @property
     def parent(self) -> GooglePath:
-        return GooglePath(self.bucket, os.path.dirname(self.blob))
+        return GooglePath(self.bucket, os.path.dirname(_strip_slash(self.blob)))
 
     def relative_to(self, other: GooglePath) -> str:
         if not isinstance(other, GooglePath):
@@ -240,7 +253,7 @@ def pathdispatch(fn: F) -> F:
     causes type checkers to complain about "no attribute register"
 
     """
-    ret: F = functools.singledispatch(fn)  # type: ignore
+    ret: F = functools.singledispatch(fn)  # type: ignore[assignment]
 
     @ret.register  # type: ignore
     def strdispatch(path: str, *args: Any, **kwargs: Any) -> Any:
@@ -254,49 +267,105 @@ def pathdispatch(fn: F) -> F:
 # ==============================
 
 
-class Stat(NamedTuple):
+class Stat:
     size: int
     mtime: float
     ctime: float
     md5: Optional[str]
     version: Optional[str]
 
-    @staticmethod
-    def from_azure(item: Mapping[str, Any]) -> Stat:
-        if "Creation-Time" in item:
-            raw_ctime = item["Creation-Time"]
-        else:
-            raw_ctime = item["x-ms-creation-time"]
-        if "x-ms-meta-blobfilemtime" in item:
-            mtime = float(item["x-ms-meta-blobfilemtime"])
-        else:
-            mtime = _azure_parse_timestamp(item["Last-Modified"])
-        return Stat(
-            size=int(item["Content-Length"]),
-            mtime=mtime,
-            ctime=_azure_parse_timestamp(raw_ctime),
-            md5=_azure_get_md5(item),
-            version=item["Etag"],
+
+class AzureStat(Stat):
+    __slots__ = (
+        "size",
+        "_mtime",
+        "_raw_mtime",
+        "_ctime",
+        "_raw_ctime",
+        "_md5",
+        "_raw_md5",
+        "version",
+    )
+
+    def __init__(self, item: Mapping[str, Any]) -> None:
+        self.size = int(item["Content-Length"])
+        self.version = item["Etag"]
+        self._raw_ctime = (
+            item["Creation-Time"] if "Creation-Time" in item else item["x-ms-creation-time"]
+        )
+        self._ctime: Optional[float] = None
+        self._raw_mtime = item["Last-Modified"]
+        self._mtime: Optional[float] = None
+        self._raw_md5 = item.get("Content-MD5")
+        self._md5: Optional[str] = ""
+
+    @property
+    def ctime(self) -> float:  # type: ignore[override]
+        if self._ctime is None:
+            self._ctime = _azure_parse_timestamp(self._raw_ctime)
+        return self._ctime
+
+    @property
+    def mtime(self) -> float:  # type: ignore[override]
+        if self._mtime is None:
+            self._mtime = _azure_parse_timestamp(self._raw_mtime)
+        return self._mtime
+
+    @property
+    def md5(self) -> Optional[str]:  # type: ignore[override]
+        if self._md5 == "":
+            self._md5 = _azure_get_md5(self._raw_md5)
+        return self._md5
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, Stat):
+            return False
+        return (
+            self.size == other.size
+            and self.mtime == other.mtime
+            and self.ctime == other.ctime
+            and self.md5 == other.md5
+            and self.version == other.version
         )
 
-    @staticmethod
-    def from_google(item: Mapping[str, Any]) -> Stat:
-        if "metadata" in item and "blobfile-mtime" in item["metadata"]:
-            mtime = float(item["metadata"]["blobfile-mtime"])
-        else:
-            mtime = _google_parse_timestamp(item["updated"])
-        return Stat(
-            size=int(item["size"]),
-            mtime=mtime,
-            ctime=_google_parse_timestamp(item["timeCreated"]),
-            md5=_google_get_md5(item),
-            version=item["generation"],
+
+class GoogleStat(Stat):
+    def __init__(self, item: Mapping[str, Any]) -> None:
+        self.size = int(item["size"])
+        self.mtime = _google_parse_timestamp(item["updated"])
+        self.ctime = _google_parse_timestamp(item["timeCreated"])
+        self.md5 = _google_get_md5(item)
+        self.version = item["generation"]
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, Stat):
+            return False
+        return (
+            self.size == other.size
+            and self.mtime == other.mtime
+            and self.ctime == other.ctime
+            and self.md5 == other.md5
+            and self.version == other.version
         )
 
-    @staticmethod
-    def from_local(item: os.stat_result) -> Stat:
-        return Stat(
-            size=item.st_size, mtime=item.st_mtime, ctime=item.st_ctime, md5=None, version=None
+
+class LocalStat(Stat):
+    def __init__(self, item: os.stat_result) -> None:
+        self.size = item.st_size
+        self.mtime = item.st_mtime
+        self.ctime = item.st_ctime
+        self.md5 = None
+        self.version = None
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, Stat):
+            return False
+        return (
+            self.size == other.size
+            and self.mtime == other.mtime
+            and self.ctime == other.ctime
+            and self.md5 == other.md5
+            and self.version == other.version
         )
 
 
@@ -306,7 +375,7 @@ async def stat(path: Union[BasePath, str]) -> Stat:
 
 
 @stat.register  # type: ignore
-async def _azure_stat(path: AzurePath) -> Stat:
+async def _azure_stat(path: AzurePath) -> AzureStat:
     if not path.blob:
         raise FileNotFoundError(path)
     request = await azurify_request(
@@ -317,11 +386,11 @@ async def _azure_stat(path: AzurePath) -> Stat:
         )
     )
     async with request.execute() as resp:
-        return Stat.from_azure(resp.headers)
+        return AzureStat(resp.headers)
 
 
 @stat.register  # type: ignore
-async def _google_stat(path: GooglePath) -> Stat:
+async def _google_stat(path: GooglePath) -> GoogleStat:
     if not path.blob:
         raise FileNotFoundError(path)
     request = await googlify_request(
@@ -333,12 +402,12 @@ async def _google_stat(path: GooglePath) -> Stat:
     )
     async with request.execute() as resp:
         result = await resp.json()
-        return Stat.from_google(result)
+        return GoogleStat(result)
 
 
 @stat.register  # type: ignore
-async def _local_stat(path: LocalPath) -> Stat:
-    return Stat.from_local(os.stat(path))
+async def _local_stat(path: LocalPath) -> LocalStat:
+    return LocalStat(os.stat(path))
 
 
 # ==============================
@@ -392,8 +461,10 @@ async def _azure_isdir(path: AzurePath) -> bool:
                 )
             )
             async for result in it:
-                if result["Blobs"] is not None:
-                    return "BlobPrefix" in result["Blobs"] or "Blob" in result["Blobs"]
+                blobs = result.find("Blobs")
+                assert blobs is not None
+                if len(blobs):
+                    return blobs.find("Blob") is not None or blobs.find("BlobPrefix") is not None
             return False
         else:
             request = await azurify_request(
@@ -511,18 +582,15 @@ def _strip_slash(path: str) -> str:
 
 
 def _azure_parse_timestamp(text: str) -> float:
-    return datetime.datetime.strptime(
-        text.replace("GMT", "Z"), "%a, %d %b %Y %H:%M:%S %z"
-    ).timestamp()
+    # A little more than 2x faster than:
+    # strptime(text.replace("GMT", "Z"), "%a, %d %b %Y %H:%M:%S %z").timestamp()
+    return email.utils.parsedate_to_datetime(text).timestamp()
 
 
-def _azure_get_md5(metadata: Mapping[str, Any]) -> Optional[str]:
-    if "Content-MD5" in metadata:
-        b64_encoded = metadata["Content-MD5"]
-        if b64_encoded is None:
-            return None
-        return base64.b64decode(b64_encoded).hex()
-    return None
+def _azure_get_md5(content_md5: Optional[str]) -> Optional[str]:
+    if content_md5 is None:
+        return None
+    return base64.b64decode(content_md5).hex()
 
 
 def _google_parse_timestamp(text: str) -> float:
