@@ -3,8 +3,6 @@ from __future__ import annotations
 import itertools
 from typing import Any, Iterator, Optional, Tuple, Union
 
-import aiohttp
-
 from .boost import (
     BoostExecutor,
     BoostUnderlying,
@@ -13,7 +11,7 @@ from .boost import (
 )
 from .globals import config
 from .path import AzurePath, BasePath, CloudPath, GooglePath, LocalPath, getsize, pathdispatch
-from .request import Request, azurify_request, googlify_request
+from .request import Request, azurify_request, execute_retrying_read, googlify_request
 
 ByteRange = Tuple[int, int]
 OptByteRange = Tuple[Optional[int], Optional[int]]
@@ -24,7 +22,7 @@ OptByteRange = Tuple[Optional[int], Optional[int]]
 
 
 @pathdispatch
-async def read_byte_range(path: Union[CloudPath, str], byte_range: OptByteRange) -> bytes:
+async def read_byte_range(path: Union[BasePath, str], byte_range: OptByteRange) -> bytes:
     """Read the content of ``path`` in the given byte range.
 
     :param path: The path to read from.
@@ -37,7 +35,7 @@ async def read_byte_range(path: Union[CloudPath, str], byte_range: OptByteRange)
 
 @read_byte_range.register  # type: ignore
 async def _azure_read_byte_range(path: AzurePath, byte_range: OptByteRange) -> bytes:
-    range_str = _byte_range_to_str(byte_range)
+    range_str = byte_range_to_str(byte_range)
     range_header = {"Range": range_str} if range_str is not None else {}
     success_codes = (206,) if range_header else (200,)
     request = await azurify_request(
@@ -46,14 +44,15 @@ async def _azure_read_byte_range(path: AzurePath, byte_range: OptByteRange) -> b
             url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
             headers=range_header,
             success_codes=success_codes,
+            failure_exceptions={404: FileNotFoundError(path)},
         )
     )
-    return await _execute_read_retrying_payload_error(request)
+    return await execute_retrying_read(request)
 
 
 @read_byte_range.register  # type: ignore
 async def _google_read_byte_range(path: GooglePath, byte_range: OptByteRange) -> bytes:
-    range_str = _byte_range_to_str(byte_range)
+    range_str = byte_range_to_str(byte_range)
     range_header = {"Range": range_str} if range_str is not None else {}
     success_codes = (206,) if range_header else (200,)
     request = await googlify_request(
@@ -63,9 +62,20 @@ async def _google_read_byte_range(path: GooglePath, byte_range: OptByteRange) ->
             params=dict(alt="media"),
             headers=range_header,
             success_codes=success_codes,
+            failure_exceptions={404: FileNotFoundError(path)},
         )
     )
-    return await _execute_read_retrying_payload_error(request)
+    return await execute_retrying_read(request)
+
+
+@read_byte_range.register  # type: ignore
+async def _local_read_byte_range(path: LocalPath, byte_range: OptByteRange) -> bytes:
+    with open(path, "rb") as f:
+        start, end = byte_range
+        if start is None:
+            return f.read() if end is None else f.read(end)
+        f.seek(start)
+        return f.read() if end is None else f.read(end - start)
 
 
 # ==============================
@@ -182,7 +192,7 @@ async def read_stream_unordered(
 # ==============================
 
 
-def _byte_range_to_str(byte_range: OptByteRange) -> Optional[str]:
+def byte_range_to_str(byte_range: OptByteRange) -> Optional[str]:
     # https://docs.microsoft.com/en-us/rest/api/storageservices/specifying-the-range-header-for-blob-service-operations
     # https://cloud.google.com/storage/docs/xml-api/get-object-download
     # oddly range requests are not mentioned in JSON API, only in the XML API
@@ -197,17 +207,3 @@ def _byte_range_to_str(byte_range: OptByteRange) -> Optional[str]:
         # This form is not supported by Azure
         return f"bytes=-{-int(end)}"
     return None
-
-
-async def _execute_read_retrying_payload_error(request: Request) -> bytes:
-    # work around https://github.com/aio-libs/aiohttp/issues/4581 /
-    # https://github.com/aio-libs/aiohttp/issues/3904#issuecomment-737094416
-    RETRIES = 3
-    for i in range(RETRIES):
-        try:
-            async with request.execute() as resp:
-                return await resp.read()
-        except aiohttp.ClientPayloadError:
-            if i >= RETRIES - 1:
-                raise
-    raise AssertionError
