@@ -1,5 +1,7 @@
 import asyncio
 import os
+import re
+import sys
 from dataclasses import dataclass
 from typing import AsyncIterator, Iterator, List, Optional, Tuple, Union
 
@@ -10,22 +12,24 @@ from .listing import DirEntry, scantree
 from .path import BasePath, LocalPath, Stat
 
 
-@dataclass
+@dataclass(frozen=True)
 class Action:
     relpath: str
 
 
-@dataclass
+@dataclass(frozen=True)
 class CopyAction(Action):
     size: Optional[int]
 
 
-@dataclass
+@dataclass(frozen=True)
 class DeleteAction(Action):
     pass
 
 
-async def sync_action_iterator(src: BasePath, dst: BasePath) -> Iterator[Action]:
+async def sync_action_iterator(
+    src: BasePath, dst: BasePath, exclude: Optional[str] = None
+) -> Iterator[Action]:
     """Yields the actions to take to sync the tree rooted at ``src`` to ``dst``.
 
     :param src: The root of the tree to copy from.
@@ -37,45 +41,57 @@ async def sync_action_iterator(src: BasePath, dst: BasePath) -> Iterator[Action]
     if isinstance(dst, LocalPath):
         dst = LocalPath(os.path.abspath(dst.path))
 
+    try:
+        exclude_pattern = re.compile(exclude) if exclude else None
+    except re.error as e:
+        raise ValueError(
+            f"Failed to compile exclude pattern {repr(exclude)}: {e}\n"
+            "Hint: exclude patterns should be Python regular expressions, not globs."
+        ) from None
+
     async def collect_tree(tree: BasePath) -> List[Tuple[str, DirEntry]]:
         try:
             # Note that if you create marker files to mark directories, scantree will return
             # those. Filter by is_file to avoid syncing these files that represent directories.
-            return [(p.path.relative_to(tree), p) async for p in scantree(tree) if p.is_file]
+            files = [(p.path.relative_to(tree), p) async for p in scantree(tree) if p.is_file]
+            return [p for p in files if exclude_pattern is None or not exclude_pattern.search(p[0])]
         except FileNotFoundError:
             return []
 
-    # We currently don't attempt to stream these actions, for fear that the list operations might
-    # start to reflect our changes, causing weird things to happen.
+    # We block on tree collection, rather than streaming actions, for fear that the list
+    # operations might start to reflect our changes, causing weird things to happen.
     src_files, dst_files = await asyncio.gather(collect_tree(src), collect_tree(dst))
+    return sync_files_action_iterator(src_files, dst_files)
+
+
+def sync_files_action_iterator(
+    src_files: List[Tuple[str, DirEntry]], dst_files: List[Tuple[str, DirEntry]]
+) -> Iterator[Action]:
     src_files.sort(key=lambda p: p[0])
     dst_files.sort(key=lambda p: p[0])
 
-    def action_iterator() -> Iterator[Action]:
-        i = 0
-        j = 0
-        while i < len(src_files) or j < len(dst_files):
-            if i < len(src_files) and (j == len(dst_files) or src_files[i][0] < dst_files[j][0]):
+    i = 0
+    j = 0
+    while i < len(src_files) or j < len(dst_files):
+        if i < len(src_files) and (j == len(dst_files) or src_files[i][0] < dst_files[j][0]):
+            src_stat = src_files[i][1].stat
+            size = src_stat.size if src_stat else None
+            yield CopyAction(src_files[i][0], size)
+            i += 1
+            continue
+        if j < len(dst_files) and (i == len(src_files) or src_files[i][0] > dst_files[j][0]):
+            yield DeleteAction(dst_files[j][0])
+            j += 1
+            continue
+        if src_files[i][0] == dst_files[j][0]:
+            if should_copy(src_files[i][1].stat, dst_files[j][1].stat):
                 src_stat = src_files[i][1].stat
                 size = src_stat.size if src_stat else None
                 yield CopyAction(src_files[i][0], size)
-                i += 1
-                continue
-            if j < len(dst_files) and (i == len(src_files) or src_files[i][0] > dst_files[j][0]):
-                yield DeleteAction(dst_files[j][0])
-                j += 1
-                continue
-            if src_files[i][0] == dst_files[j][0]:
-                if _should_copy(src_files[i][1].stat, dst_files[j][1].stat):
-                    src_stat = src_files[i][1].stat
-                    size = src_stat.size if src_stat else None
-                    yield CopyAction(src_files[i][0], size)
-                i += 1
-                j += 1
-                continue
-            raise AssertionError
-
-    return action_iterator()
+            i += 1
+            j += 1
+            continue
+        raise AssertionError
 
 
 # ==============================
@@ -88,6 +104,7 @@ async def sync(
     dst: Union[str, BasePath],
     executor: BoostExecutor,
     delete: bool = False,
+    exclude: Optional[str] = None,
 ) -> AsyncIterator[BasePath]:
     """Syncs the tree rooted at ``src`` to ``dst``.
 
@@ -107,11 +124,19 @@ async def sync(
     if src_obj.is_relative_to(dst_obj) or dst_obj.is_relative_to(src_obj):
         raise ValueError("Cannot sync overlapping directories")
 
-    async def copy_wrapper(relpath: str, size: Optional[int]) -> BasePath:
+    async def copy_wrapper(relpath: str, size: Optional[int]) -> Optional[BasePath]:
         src_file = src_obj / relpath
         dst_file = dst_obj / relpath
-        await copyfile(src_file, dst_file, executor, size=size, overwrite=True)
-        return src_file
+        try:
+            await copyfile(src_file, dst_file, executor, size=size, overwrite=True)
+            return src_file
+        except FileNotFoundError as e:
+            print(
+                "[boostedblob] File disappeared while syncing, ignoring. Likely due to "
+                f"concurrent deletion: {e}",
+                file=sys.stderr,
+            )
+        return None
 
     async def delete_wrapper(relpath: str) -> BasePath:
         dst_file = dst_obj / relpath
@@ -126,13 +151,15 @@ async def sync(
                 return await delete_wrapper(action.relpath)
         return None
 
-    actions = executor.map_unordered(action_wrapper, await sync_action_iterator(src_obj, dst_obj))
+    actions = executor.map_unordered(
+        action_wrapper, await sync_action_iterator(src_obj, dst_obj, exclude=exclude)
+    )
     async for path in actions:
         if path is not None:
             yield path
 
 
-def _should_copy(src_stat: Optional[Stat], dst_stat: Optional[Stat]) -> bool:
+def should_copy(src_stat: Optional[Stat], dst_stat: Optional[Stat]) -> bool:
     if src_stat is None and dst_stat is None:
         return False
     if src_stat is None or dst_stat is None:
